#include <unistd.h>
#include <stdio.h>
#include <ctype.h>


/* This must be included in order
   to have access to the ME-9000
   specific definitions */
#include <me9000mio.h>
#include <libme9000mio.h>

static void pCntIrqProc(void *pCntIrqContext){
	unsigned int board = 0;
	unsigned long count = 0;

	board = * ((int *) pCntIrqContext);
	me9000mioCntGetIrqCount(board, &count);
	printf("Interrupt Count = %lu\n\n", count);
}


void usage(void){
	printf("cnt - Program to test the counter on a ME-9000\n\n");
	printf("Usage: cnt [arguments]\n\n");
	printf("Arguments:\n\n");
	printf("-h                     Print this help and exit.\n");
	printf("-b <number>            Use board <number> (Default is 0).\n");
	printf("-t                     If set the counter will be not retriggerd.\n");
	printf("-e                     If set the counter will cause an external event.\n");
	printf("-c                     If set the external clock input will be enabled.\n");
	printf("-v <value>             Write <value> to counter (Default is 60000).\n");
	printf("-r                     If set the value is read from the counter.\n");
}



int main(int argc, char *argv[]){
	int err;
	char errStr[256] = {0};
	unsigned int board = 0;
	int retrigger = ME9000MIO_CNT_RETRIGGER_ENABLE;
	int ext_event = ME9000MIO_CNT_EVENT_DISABLE;
	int ext_clock = ME9000MIO_CNT_CLOCK_INTERNAL;
	unsigned short value = 60000;
	int read = 0;
	int c = 0;

	/* Parse the command line arguments */
	while((c = getopt(argc, argv, "htecrb:v:")) != -1){
		switch(c){
			case 'h':
				usage();
				exit(0);
			case 'b':
				board = atoi(optarg);
				break;
			case 't':
				retrigger = ME9000MIO_CNT_RETRIGGER_DISABLE;
				break;
			case 'e':
				ext_event = ME9000MIO_CNT_EVENT_ENABLE;
				break;
			case 'c':
				ext_clock = ME9000MIO_CNT_CLOCK_EXTERNAL;
				break;
			case 'v':
				value = atoi(optarg);
				break;
			case 'r':
				read = 1;
				break;
			default:
				usage();
				exit(2);
		}
	}

	if(optind != argc){
		fprintf(stderr, "%s: No non option arguments are supported.\n", argv[0]);
		usage();
		exit(2);
	}

	printf("Open counter on board %d.\n", board);
	err = me9000mioCntOpen(board);
	if(err){
		me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
		fprintf(stderr, "ERROR: %s.\n", errStr);
		exit(1);
	}

	if(read){
		err = me9000mioCntRead(board, &value);
		if(err){
			me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
			fprintf(stderr, "ERROR: %s.\n", errStr);
			exit(1);
		}
		printf("Read %d.\n", value);
	}
	else{
		printf("Stop the counter.\n");
		err = me9000mioCntStop(board);
		if(err){
			me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
			fprintf(stderr, "ERROR: %s.\n", errStr);
			exit(1);
		}

		printf("Write %d to counter.\n", value);
		err = me9000mioCntWrite(board, value);
		if(err){
			me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
			fprintf(stderr, "ERROR: %s.\n", errStr);
			exit(1);
		}

		printf("Configure counter.\n");
		err = me9000mioCntConfig(
				board,
				retrigger,
				ext_event,
				ext_clock,
				ME9000MIO_CNT_INT_ENABLE,
				pCntIrqProc,
				&board);
		if(err){
			me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
			fprintf(stderr, "ERROR: %s.\n", errStr);
			exit(1);
		}

		printf("Start counter.\n");
		err = me9000mioCntStart(board);
		if(err){
			me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
			fprintf(stderr, "ERROR: %s.\n", errStr);
			exit(1);
		}

		printf("Type 'CTRL-D' to finish the test program\n");
		while(getchar() != EOF);
	}

	printf("Stop counter.\n");
	err = me9000mioCntStop(board);
	if(err){
		me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
		fprintf(stderr, "ERROR: %s.\n", errStr);
		exit(1);
	}

	printf("Close counter on board %d.\n", board);
	err = me9000mioCntClose(board);
	if(err){
		me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
		fprintf(stderr, "ERROR: %s.\n", errStr);
		exit(1);
	}

	exit(0);
}
