
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>


/* This must be included in order
   to have access to the ME-9000
   specific definitions */
#include <libme9000mio.h>



void usage(void){
	printf("dioBit - Program to test digital I/O\n\n");
	printf("Usage: dioBit [arguments]\n\n");
	printf("Arguments:\n\n");
	printf("-h                     Print this help and exit.\n");
	printf("-b <board number>      Use <board number> board (Default is 0).\n");
	printf("-l <line number>       Use <line number> line (Default is 0).\n");
	printf("-i                     Invert digital input line.\n");
	printf("-t                     Switch digital output line to high impedance.\n");
	printf("-w <value>             Configure corresponding line for writing and write\n");
	printf("                       <value> to it. 0 is false else is true.\n");
	printf("                       If this option is omitted the line is read.\n");
}



int main(int argc, char *argv[]){
	int err;
	char errStr[256] = {0};
	int c;
	unsigned int board = 0;
	unsigned int bit = 0;
	int dir = ME9000MIO_DIO_INPUT;
	int inv = ME9000MIO_DIO_INV_DISABLE;
	int tristate = 0;
	int value = 0;

	/* Parse the command line arguments */
	while((c = getopt(argc, argv, "hb:w:itl:")) != -1){
		switch(c){
			case 'h':
				usage();
				exit(0);
			case 'b':
				board = atoi(optarg);
				break;
			case 'w':
				dir = ME9000MIO_DIO_OUTPUT;
				value = atoi(optarg);
				break;
			case 'i':
				inv = ME9000MIO_DIO_INV_ENABLE;
				break;
			case 't':
				tristate = ME9000MIO_DIO_INV_ENABLE;
				break;
			case 'l':
				bit = atoi(optarg);
				break;
			default:
				usage();
				exit(2);
		}
	}

	if(optind != argc){
		fprintf(stderr, "%s: No non option arguments are supported.\n", argv[0]);
		usage();
		exit(2);
	}

	printf("Open digital I/O on board %u.\n", board);
	err = me9000mioDIOOpen(board);
	if(err){
		me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
		fprintf(stderr, "ERROR: %s.\n", errStr);
		exit(1);
	}

	if(dir == ME9000MIO_DIO_OUTPUT){ // Output selected
		printf("Configure line %u as output line.\n", bit);
		err = me9000mioDIOConfig(
				board,
				bit,
				dir,
				inv,
				ME9000MIO_DIO_INT_DISABLE);
		if(err){
			me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
			fprintf(stderr, "ERROR: %s.\n", errStr);
			exit(1);
		}

		if(tristate){
			printf("Set digital output line %d to high impedance\n", bit);
			err = me9000mioDIOOutput(
					board,
					bit,
					ME9000MIO_DIO_OUTPUT_DISABLE);
			if(err){
				me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
				fprintf(stderr, "ERROR: %s.\n", errStr);
				exit(1);
			}
		}
		else {
			err = me9000mioDIOOutput(
					board,
					bit,
					ME9000MIO_DIO_OUTPUT_ENABLE);
			if(err){
				me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
				fprintf(stderr, "ERROR: %s.\n", errStr);
				exit(1);
			}

			printf("Write 0x%X to digital line %d\n", value, bit);
			err = me9000mioDIOSetBit(board, bit, value);
			if(err){
				me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
				fprintf(stderr, "ERROR: %s.\n", errStr);
				exit(1);
			}
		}
	}
	else{ // Input selected
		printf("Configure line %u as input line.\n", bit);
		err = me9000mioDIOConfig(
				board,
				bit,
				dir,
				inv,
				ME9000MIO_DIO_INT_DISABLE);
		if(err){
			me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
			fprintf(stderr, "ERROR: %s.\n", errStr);
			exit(1);
		}

		printf("Read value from digital input line %d.\n", bit);
		err = me9000mioDIOGetBit(board, bit, &value);
		if(err){
			me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
			fprintf(stderr, "ERROR: %s.\n", errStr);
			exit(1);
		}
		printf("Value from digital input line is 0x%X.\n", value);
	}

	printf("Close digital I/O on board %u.\n", board);
	err = me9000mioDIOClose(board);
	if(err){
		me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
		fprintf(stderr, "ERROR: %s.\n", errStr);
		exit(1);
	}

	exit(0);
}
