#include <unistd.h>
#include <stdio.h>
#include <ctype.h>


/* This must be included in order
   to have access to the ME-9000
   specific definitions */
#include <libme9000mio.h>



/* Global stuff */
volatile int i = 0;
static char *context = "External interrupt occured\n";
unsigned int board = 0;


void callback(void *context){
	char *text;
	unsigned long count;

	text = context;
	printf("%s", text);
	me9000mioDIOGetIrqCount(board, &count);
	printf("Interrupt count = %lu\n", count);
	i++;
}


void usage(void){
	printf("DIOIrq - Program to test the external interrupt\n\n");
	printf("Usage: extIrq [arguments]\n\n");
	printf("Arguments:\n\n");
	printf("-h                     Print this help and exit.\n");
	printf("-b <board number>      Use <board number> board (Default is 0).\n");
	printf("-l <line number>       Enable external interrupt on line <line number> (Default 0).\n");
	printf("-i                     If set the input line is inverted.\n");
	printf("-c <interrupt count>   Wait for <interrupt count> interrupts (Default 8).\n");
}



int main(int argc, char *argv[]){
	int err;
	char errStr[256] = {0};
	int c;
	int line = 0;
	int invert = ME9000MIO_DIO_INV_DISABLE;
	int intCount = 8;

	/* Parse the command line arguments */
	while((c = getopt(argc, argv, "hib:l:c:")) != -1){
		switch(c){
			case 'h':
				usage();
				exit(0);
			case 'b':
				board = atoi(optarg);
				break;
			case 'l':
				line = atoi(optarg);
				break;
			case 'i':
				invert = ME9000MIO_DIO_INV_ENABLE;
				break;
			case 'c':
				intCount = atoi(optarg);
				break;
			default:
				usage();
				exit(2);
		}
	}

	if(optind != argc){
		fprintf(stderr, "%s: No non option arguments are supported.\n", argv[0]);
		usage();
		exit(2);
	}

	/*-------------------------------------------------------------------------
	  Open the DIO
	  -----------------------------------------------------------------------*/

	printf("Open external interrupt on board %d\n", board);
	err = me9000mioDIOOpen(board);
	if(err){
		me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
		fprintf(stderr, "ERROR: %s.\n", errStr);
		exit(1);
	}

	/*-------------------------------------------------------------------------
	  Enable external interrupt
	  -----------------------------------------------------------------------*/

	printf("Install callback routine\n");
	err = me9000mioDIOSetCallback(
			board,
			callback,
			context);
	if(err){
		me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
		fprintf(stderr, "ERROR: %s.\n", errStr);
		exit(1);
	}

	printf("Enable external interrupt on line %d\n", line);
	err = me9000mioDIOConfig(
			board,
			line,
			ME9000MIO_DIO_INPUT,
			invert,
			ME9000MIO_DIO_INT_ENABLE);
	if(err){
		me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
		fprintf(stderr, "ERROR: %s.\n", errStr);
		exit(1);
	}

	while(i < intCount);

	/*-------------------------------------------------------------------------
	  Disable external interrupt
	  -----------------------------------------------------------------------*/

	printf("Disable external interrupt on line %d\n", line);
	err = me9000mioDIOConfig(
			board,
			line,
			ME9000MIO_DIO_INPUT,
			invert,
			ME9000MIO_DIO_INT_DISABLE);
	if(err){
		me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
		fprintf(stderr, "ERROR: %s.\n", errStr);
		exit(1);
	}

	/*-------------------------------------------------------------------------
	  Close the external interrupt
	  -----------------------------------------------------------------------*/

	printf("Close external interrupt on board %d.\n", board);
	err = me9000mioDIOClose(board);
	if(err){
		me9000mioErrorGetMessage(err, errStr, sizeof(errStr));
		fprintf(stderr, "ERROR: %s.\n", errStr);
		exit(1);
	}

	return 0;
}
