/* API for DIO and Counter on the Meilhaus ME-9000 board family.
 * =============================================================
 *
 *  Copyright (C) 2004 Meilhaus Electronic GmbH (support@meilhaus.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Author:	Guenter Gebhardt	<g.gebhardt@meilhaus.de>
 */


#include <libme9000mio.h>
#include <libme9000mio_error.h>


/* Init and exit of the shared object */
void _init(void);
void _fini(void);


#define ME9000MIO_DLL_VERSION 0x10000


/* Global error number */
static int my_errno = 0;

/* Stores the system error message after a ERROR_NO_SYSTEM */
static char sysErrMsg[256];

/* Flag for default error function */
static int errorDefaultProcFlag = 0;

/* Pointer to user defined error function */
static ME9000MIO_P_ERROR_PROC errorUserProcPtr = NULL;

/* Holds the number of the detected boards */
static int board_count = 0;


typedef struct me9000DIOContext{
    int fd;
    ME9000MIO_P_DIO_PROC pCallback;
    void *pCallbackContext;
} me9000mioDIOContext_t;


typedef struct me9000CntContext{
    int fd;
    ME9000MIO_P_CNT_PROC pCallback;
    void *pCallbackContext;
} me9000mioCntContext_t;


typedef struct me9000mioBoardContext {
	unsigned short usBoardVersion;
	unsigned long ulSerialNumber;
    me9000mioDIOContext_t dioContext;
    me9000mioCntContext_t cntContext;
} me9000mioBoardContext_t;

static me9000mioBoardContext_t boardContextVec[4];


static char *errorTable[1024] = {
	ERROR_STR_SUCCESS,
	ERROR_STR_UNSPECIFIED,
	ERROR_STR_INVALID_BOARD,
	ERROR_STR_NOT_OPEN,
	ERROR_STR_INVALID_CHANNEL,
	ERROR_STR_INVALID_BIT,
	ERROR_STR_INVALID_COUNTER,
	ERROR_STR_INVALID_COUNTER_MODE,
	ERROR_STR_BUSY,
	ERROR_STR_INVALID_OUTPUT_MODE,
	ERROR_STR_INVALID_ERROR_NO,
	ERROR_STR_ERROR_BUFFER_TO_LITTLE,
	ERROR_STR_SYSTEM
};



/*=============================================================================
  Internal error functions
  ===========================================================================*/

static int storeSystemError(int error);
static void errorDefaultProc(char *function, int iErrorCode);


/*=============================================================================
  Signal handler
  ===========================================================================*/

static void irqHandler(int sig);


/*=============================================================================
  Internal DIO functions
  ===========================================================================*/

static int DIOCheckBoard(unsigned int uiBoardNumber);
static int DIOCheckBoardFD(unsigned int uiBoardNumber);
static int DIOOpen(me9000mioDIOContext_t *dioContext, unsigned int uiBoardNumber);
static int DIOClose(me9000mioDIOContext_t *dioContext);
static int DIOConfig(me9000mioDIOContext_t *dioContext, unsigned int uiBitNumber, int iDirection, int iInvert, int iInterrupt);
static int DIOSetCallback(me9000mioDIOContext_t *dioConfig, ME9000MIO_P_DIO_PROC pDIOIrqProc, void *pDIOIrqContext);
static int DIOGetBit(me9000mioDIOContext_t *dioConfig, unsigned int uiBitNumber, int *piBitValue);
static int DIOGetByte(me9000mioDIOContext_t *dioConfig, unsigned char *pucByteValue);
static int DIOSetBit(me9000mioDIOContext_t *dioConfig, unsigned int uiBitNumber, int iBitValue);
static int DIOSetByte(me9000mioDIOContext_t *dioContext, unsigned char ucByteValue);
static int DIOOutput(me9000mioDIOContext_t *dioContext, unsigned int uiBitNumber, int iMode);
static int DIOGetIrqCount(me9000mioDIOContext_t *dioConfig, unsigned long *pulCount);
static int DIOReset(me9000mioDIOContext_t *dioContext);


/*=============================================================================
  Internal counter functions
  ===========================================================================*/

static int cntCheckBoard(unsigned int uiBoardNumber);
static int cntCheckBoardFD(unsigned int uiBoardNumber);
static int cntOpen(me9000mioCntContext_t *cntContext, unsigned int uiBoardNumber);
static int cntClose(me9000mioCntContext_t *cntContext);
static int cntConfig(
	   	me9000mioCntContext_t *cntContext,
		int iRetrigger,
		int iExtEvent,
		int iInterrupt,
		int iExtClock,
		ME9000MIO_P_CNT_PROC pCntIrqProc,
		void *pCntIrqContext);
static int cntWrite(me9000mioCntContext_t *cntContext, unsigned short usValue);
static int cntRead(me9000mioCntContext_t *cntContext, unsigned short *pusValue);
static int cntStart(me9000mioCntContext_t *cntContext);
static int cntStop(me9000mioCntContext_t *cntContext);
static int cntGetIrqCount(me9000mioCntContext_t *cntContext, unsigned long *pulCount);
static int cntReset(me9000mioCntContext_t *cntContext);


/*=============================================================================
  Initialization and cleanup of shared object
  ===========================================================================*/

void _init(void){
	int fd;
	char path[32];
	me9000mio_user_info_t info;

	/* Detect number of board available and get the board information */
	for(board_count = 0; board_count < 4; board_count++){
		sprintf(path, "/dev/me9000mio_%d_dio", board_count);
		fd = open(path, O_RDWR);
		fprintf(stderr, "Open '%s' %s.\n", path, (fd >= 0)?"SUCCESSED":"FAILED");
		if(fd < 0)
		{
			break;
		}

		ioctl(fd, ME9000MIO_GET_USER_INFO, &info);

		close(fd);

		/* Init board context */
		boardContextVec[board_count].usBoardVersion = info.device_id;
		boardContextVec[board_count].ulSerialNumber = info.serial_no;

		/* Init dio context */
		boardContextVec[board_count].dioContext.fd = -1;
		boardContextVec[board_count].dioContext.pCallback = NULL;
		boardContextVec[board_count].dioContext.pCallbackContext = NULL;

		/* Init cnt context */
		boardContextVec[board_count].cntContext.fd = -1;
		boardContextVec[board_count].cntContext.pCallback = NULL;
		boardContextVec[board_count].cntContext.pCallbackContext = NULL;
	    fprintf(stderr, "Board %d add to context.\n", board_count);
	}
	fprintf(stderr, "Total numbers of boards: %d.\n", board_count);

}


/*===========================================================================
  Error handling functions
  =========================================================================*/

int me9000mioErrorGetMessage(
		int iErrorCode,
		char *pcBuffer,
		unsigned int uiBufferSize){
	if(iErrorCode > ERROR_NO_SYSTEM || iErrorCode < ERROR_NO_SUCCESS){
		my_errno = ERROR_NO_INVALID_ERROR_NO;
		strncpy(pcBuffer, errorTable[ERROR_NO_INVALID_ERROR_NO], uiBufferSize);
		return ME9000MIO_NO_ERROR;
	}

	if(iErrorCode == ERROR_NO_SYSTEM)
		strncpy(pcBuffer, sysErrMsg, uiBufferSize);
	else
		strncpy(pcBuffer, errorTable[iErrorCode], uiBufferSize);

	return ME9000MIO_NO_ERROR;
}


int me9000mioErrorGetLastMessage(char *pcBuffer, unsigned int uiBufferSize){
    return me9000mioErrorGetMessage(my_errno, pcBuffer, uiBufferSize);
}


int me9000mioErrorSetDefaultProc(int iDefaultProcStatus){
    errorDefaultProcFlag = iDefaultProcStatus;
    return ME9000MIO_NO_ERROR;
}


int me9000mioErrorSetUserProc(ME9000MIO_P_ERROR_PROC pErrorProc){
    errorUserProcPtr = pErrorProc;
    return ME9000MIO_NO_ERROR;
}


int storeSystemError(int error){
    strcpy(sysErrMsg, strerror(error));
    my_errno = ERROR_NO_SYSTEM;
    return ERROR_NO_SYSTEM;
}


void errorDefaultProc(char *function, int iErrorCode){
    char msg[64] = {0};
    me9000mioErrorGetMessage(iErrorCode, msg, sizeof(msg));
    fprintf(stderr, "In function %s: %s (%d)\n", function, msg, iErrorCode);
}



/*=============================================================================
  General purpose functions
  ===========================================================================*/

int me9000mioGetDLLVersion(unsigned long *pulVersion){
	*pulVersion = 0x010000; // Version xx.xx.xx
	return ME9000MIO_NO_ERROR;
}


int me9000mioGetDriverVersion(unsigned long *pulVersion){
	*pulVersion = 0x010000; // Version xx.xx.xx
	return ME9000MIO_NO_ERROR;
}


int me9000mioGetBoardCount(unsigned int *puiCount){
	*puiCount = board_count;
    return ME9000MIO_NO_ERROR;
}


int me9000mioGetBoardVersion(unsigned int uiBoardNumber, unsigned short *pusVersion){
	int err;

	err = DIOCheckBoard(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioGetBoardVersion()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioGetBoardVersion()", err);
		return err;
	}

	*pusVersion = boardContextVec[uiBoardNumber].usBoardVersion;
	return ME9000MIO_NO_ERROR;
}


int me9000mioGetSerialNumber(unsigned int uiBoardNumber, unsigned long *pulSerialNumber){
	int err;

	err = DIOCheckBoard(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioGetSerialNumber()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioGetSerialNumber()", err);
		return err;
	}

	*pulSerialNumber = boardContextVec[uiBoardNumber].ulSerialNumber;
	return ME9000MIO_NO_ERROR;
}


/*=============================================================================
  Digital I/O functions
  ===========================================================================*/

int DIOCheckBoard(unsigned int uiBoardNumber){
    fprintf(stderr, "In function %s: board_count(%d)\n", __FUNCTION__, board_count);

	if(uiBoardNumber >= board_count){
		errno = EINVAL;
		my_errno = ERROR_NO_INVALID_BOARD;
		return ERROR_NO_INVALID_BOARD;
	}
	return ME9000MIO_NO_ERROR;
}


int DIOCheckBoardFD(unsigned int uiBoardNumber){
	int err;

	err = DIOCheckBoard(uiBoardNumber);
	if(err) return err;

	if(boardContextVec[uiBoardNumber].dioContext.fd < 0){
		errno = EINVAL;
		my_errno = ERROR_NO_NOT_OPEN;
		return ERROR_NO_NOT_OPEN;
	}

	return ME9000MIO_NO_ERROR;
}


int me9000mioDIOOpen(unsigned int uiBoardNumber){
	int err;

	err = DIOCheckBoard(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOOpen()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOOpen()", err);
		return err;
	}

	if(boardContextVec[uiBoardNumber].dioContext.fd != -1){
		return ME9000MIO_NO_ERROR; // Already open
	}

	err = DIOOpen(&boardContextVec[uiBoardNumber].dioContext, uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOOpen()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOOpen()", err);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int DIOOpen(me9000mioDIOContext_t *dioContext, unsigned int uiBoardNumber){
	char str[64];
	int err;

	snprintf(str, sizeof(str), "/dev/me9000mio_%u_dio", uiBoardNumber);

	err = open(str, O_RDWR);
	if(err < 0) return storeSystemError(errno);
	dioContext->fd = err;
	return ME9000MIO_NO_ERROR;
}


int me9000mioDIOClose(unsigned int uiBoardNumber){
	int err;

	err = DIOCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOClose()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOClose()", err);
		return err;
	}

	err = DIOClose(&boardContextVec[uiBoardNumber].dioContext);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOClose()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOClose()", err);
		return err;
	}

	return ME9000MIO_NO_ERROR;
}


int DIOClose(me9000mioDIOContext_t *dioContext){
	int err;

	err = close(dioContext->fd);
	if(err) return storeSystemError(errno);
	dioContext->fd = -1;
	return ME9000MIO_NO_ERROR;
}


int me9000mioDIOConfig(
		unsigned int uiBoardNumber,
		unsigned int uiBitNumber,
		int iDirection,
		int iInvert,
		int iInterrupt){
	int err;

	err = DIOCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOConfig()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOConfig()", err);
		return err;
	}

	err = DIOConfig(
			&boardContextVec[uiBoardNumber].dioContext,
			uiBitNumber,
			iDirection,
			iInvert,
			iInterrupt);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOConfig()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOConfig()", err);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int DIOConfig(
		me9000mioDIOContext_t *dioContext,
		unsigned int uiBitNumber,
		int iDirection,
		int iInvert,
		int iInterrupt){
	int err;
	me9000mio_dio_config_t config;

	config.line = uiBitNumber;
	config.dir = iDirection;
	config.invert = iInvert;
	config.interrupt = iInterrupt;

    err = ioctl(dioContext->fd, ME9000MIO_DIO_CONFIG, &config);
    if(err) return storeSystemError(errno);

    return ME9000MIO_NO_ERROR;
}


int me9000mioDIOReset(unsigned int uiBoardNumber){
	int err;

	err = DIOCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOReset()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOReset()", err);
		return err;
	}

	err = DIOReset(&boardContextVec[uiBoardNumber].dioContext);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOReset()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOReset()", err);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int DIOReset(me9000mioDIOContext_t *dioContext){
	int err;
	int i;

	for (i = 0; i < 8; i++){
		err = DIOConfig(
				dioContext,
				i,
				ME9000MIO_DIO_INPUT,
				ME9000MIO_DIO_INV_DISABLE,
				ME9000MIO_DIO_INT_DISABLE);
		if (err) return err;

		err = DIOOutput(
				dioContext,
				i,
				ME9000MIO_DIO_OUTPUT_ENABLE);
		if (err) return err;
	}

    return ME9000MIO_NO_ERROR;
}


int me9000mioDIOSetCallback(
		unsigned int uiBoardNumber,
		ME9000MIO_P_DIO_PROC pDIOIrqProc,
		void *pDIOIrqContext){
	int err;

	err = DIOCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOSetCallback()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOSetCallback()", err);
		return err;
	}

	err = DIOSetCallback(
			&boardContextVec[uiBoardNumber].dioContext,
			pDIOIrqProc,
			pDIOIrqContext);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOSetCallback()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOSetCallback()", err);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int DIOSetCallback(
		me9000mioDIOContext_t *dio_context,
		ME9000MIO_P_DIO_PROC pDIOIrqProc,
		void *pDIOIrqContext){
	int err;
	int oflags;

	if(pDIOIrqProc){
		dio_context->pCallback = pDIOIrqProc;
		dio_context->pCallbackContext = pDIOIrqContext;

		/* Install signal handler */
		if(signal(SIGIO, irqHandler) == SIG_ERR)
			return storeSystemError(errno);

		/* Register process as owner of the path */
		err = fcntl(dio_context->fd, F_SETOWN, getpid());
		if(err == -1)
			return storeSystemError(errno);

		/* Ask for the flags set in the path */
		oflags = fcntl(dio_context->fd, F_GETFL);
		if(oflags == -1)
			return storeSystemError(errno);

		/* Set the current process to the asynchronous queue of the driver */
		err = fcntl(dio_context->fd, F_SETFL, oflags | FASYNC);
		if(err == -1)
			return storeSystemError(errno);
	}
	else{
		dio_context->pCallback = NULL;
		dio_context->pCallbackContext = NULL;

		/* Uninstall signal handler */
		if(signal(SIGIO, SIG_DFL) == SIG_ERR)
			return storeSystemError(errno);

		/* Register process as owner of the path */
		err = fcntl(dio_context->fd, F_SETOWN, getpid());
		if(err == -1)
			return storeSystemError(errno);

		/* Ask for the flags set in the path */
		oflags = fcntl(dio_context->fd, F_GETFL);
		if(oflags == -1)
			return storeSystemError(errno);

		/* Unset the current process to the asuncronous queue of the driver */
		fcntl(dio_context->fd, F_SETFL, oflags & ~FASYNC);
		if(err == -1)
			return storeSystemError(errno);
	}

	return ME9000MIO_NO_ERROR;
}


int me9000mioDIOGetBit(
		unsigned int uiBoardNumber,
		unsigned int uiBitNumber,
		int *piBitValue){
	int err;

	err = DIOCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOGetBit()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOGetBit()", err);
		return err;
	}

	err = DIOGetBit(
			&boardContextVec[uiBoardNumber].dioContext,
			uiBitNumber,
			piBitValue);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOGetBit()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOGetBit()", err);
		return err;
	}

	return ME9000MIO_NO_ERROR;
}


int DIOGetBit(
		me9000mioDIOContext_t *dioConfig,
		unsigned int uiBitNumber,
		int *piBitValue){
	me9000mio_dio_line_t cmd;
	int err;

	cmd.line = uiBitNumber;
	err = ioctl(dioConfig->fd, ME9000MIO_DIO_GET_LINE, &cmd);
	if(err) return storeSystemError(errno);

	*piBitValue = cmd.status;
	return ME9000MIO_NO_ERROR;
}


int me9000mioDIOGetByte(
		unsigned int uiBoardNumber,
		unsigned char *pucByteValue){
	int err;

	err = DIOCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOGetByte()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOGetByte()", err);
		return err;
	}

	err = DIOGetByte(&boardContextVec[uiBoardNumber].dioContext, pucByteValue);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOGetByte()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOGetByte()", err);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int DIOGetByte(
		me9000mioDIOContext_t *dioConfig,
		unsigned char *pucByteValue){
	uint8_t byte;
	int err;

	err = ioctl(dioConfig->fd, ME9000MIO_DIO_GET_BYTE, &byte);
	if(err) return storeSystemError(errno);

	*pucByteValue = byte;
	return ME9000MIO_NO_ERROR;
}


int me9000mioDIOSetBit(
		unsigned int uiBoardNumber,
		unsigned int uiBitNumber,
		int iBitValue){
	int err;

	err = DIOCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOSetBit()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOSetBit()", err);
		return err;
	}

	err = DIOSetBit(&boardContextVec[uiBoardNumber].dioContext, uiBitNumber, iBitValue);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOSetBit()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOSetBit()", err);
		return err;
	}

	return ME9000MIO_NO_ERROR;
}



int DIOSetBit(
		me9000mioDIOContext_t *dioConfig,
		unsigned int uiBitNumber,
		int iBitValue){
	me9000mio_dio_line_t cmd;
	int err;

	cmd.line = uiBitNumber;
	cmd.status = iBitValue;

	err = ioctl(dioConfig->fd, ME9000MIO_DIO_SET_LINE, &cmd);
	if(err) return storeSystemError(errno);

	return ME9000MIO_NO_ERROR;
}


int me9000mioDIOSetByte(
		unsigned int uiBoardNumber,
		unsigned char ucByteValue){
	int err;

	err = DIOCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOSetByte()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOSetByte()", err);
		return err;
	}

	err = DIOSetByte(&boardContextVec[uiBoardNumber].dioContext, ucByteValue);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOSetByte()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOSetByte()", err);
		return err;
	}

	return ME9000MIO_NO_ERROR;
}


int DIOSetByte(
		me9000mioDIOContext_t *dioContext,
		unsigned char ucByteValue){
	int err;
	uint8_t byte;

	byte = ucByteValue;

	err = ioctl(dioContext->fd, ME9000MIO_DIO_SET_BYTE, &byte);
	if(err) return storeSystemError(errno);

	return ME9000MIO_NO_ERROR;
}


int me9000mioDIOOutput(
		unsigned int uiBoardNumber,
		unsigned int uiBitNumber,
		int iOutputMode){
	int err;

	err = DIOCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOOutput()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOOutput()", err);
		return err;
	}

	err = DIOOutput(&boardContextVec[uiBoardNumber].dioContext, uiBitNumber, iOutputMode);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOOutput()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOOutput()", err);
		return err;
	}

	return ME9000MIO_NO_ERROR;
}


int DIOOutput(
		me9000mioDIOContext_t *dioContext,
		unsigned int uiBitNumber,
		int iOutputMode){
	me9000mio_dio_line_t cmd;
	int err;

	cmd.line = uiBitNumber;
	cmd.status = iOutputMode;

	err = ioctl(dioContext->fd, ME9000MIO_DIO_TRISTATE, &cmd);
	if(err) return storeSystemError(errno);

	return ME9000MIO_NO_ERROR;
}


int me9000mioDIOGetIrqCount(
		unsigned int uiBoardNumber,
		unsigned long *pulCount){
	int err;

	err = DIOCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOGetIrqCount()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOGetIrqCount()", err);
		return err;
	}

	err = DIOGetIrqCount(&boardContextVec[uiBoardNumber].dioContext, pulCount);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioDIOGetIrqCount()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioDIOGetIrqCount()", err);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int DIOGetIrqCount(
		me9000mioDIOContext_t *dioConfig,
		unsigned long *pulCount){
	int err;

	err = ioctl(dioConfig->fd, ME9000MIO_DIO_GET_IRQ_COUNT, pulCount);
	if(err) return storeSystemError(errno);

	return ME9000MIO_NO_ERROR;
}



/*===========================================================================
  Counter routines
  =========================================================================*/

int cntCheckBoard(unsigned int uiBoardNumber){
	if(uiBoardNumber >= board_count){
		errno = EINVAL;
		my_errno = ERROR_NO_INVALID_BOARD;
		return ERROR_NO_INVALID_BOARD;
	}
	return ME9000MIO_NO_ERROR;
}


int cntCheckBoardFD(unsigned int uiBoardNumber){
	int err;

	err = cntCheckBoard(uiBoardNumber);
	if(err) return err;

	if(boardContextVec[uiBoardNumber].cntContext.fd < 0){
		errno = EINVAL;
		my_errno = ERROR_NO_NOT_OPEN;
		return ERROR_NO_NOT_OPEN;
	}
	return ME9000MIO_NO_ERROR;
}


int me9000mioCntOpen(unsigned int uiBoardNumber){
	int err;

	err = cntCheckBoard(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntOpen()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntOpen()", err);
		return err;
	}

	if(boardContextVec[uiBoardNumber].cntContext.fd != -1){
		return ME9000MIO_NO_ERROR; // Already open
	}

	err = cntOpen(&boardContextVec[uiBoardNumber].cntContext, uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntOpen()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntOpen()", err);
		return err;
	}

	return ME9000MIO_NO_ERROR;
}


int cntOpen(me9000mioCntContext_t *cntContext, unsigned int uiBoardNumber){
	char str[64] = {0};
	int err;

	snprintf(str, sizeof(str), "/dev/me9000mio_%u_cnt", uiBoardNumber);

	err = open(str, O_RDWR);
	if(err < 0) return storeSystemError(errno);
	cntContext->fd = err;
	return ME9000MIO_NO_ERROR;
}


int me9000mioCntClose(unsigned int uiBoardNumber){
	int err;

	err = cntCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntClose()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntClose()", err);
		return err;
	}

	err = cntClose(&boardContextVec[uiBoardNumber].cntContext);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntClose()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntClose()", err);
		return err;
	}

	return ME9000MIO_NO_ERROR;
}


int cntClose(me9000mioCntContext_t *cntContext){
	int err;

	err = close(cntContext->fd);
	if(err) return storeSystemError(errno);

	cntContext->fd = -1;
	return ME9000MIO_NO_ERROR;
}


int me9000mioCntConfig(
		unsigned int uiBoardNumber,
		int iRetrigger,
		int iExtEvent,
		int iClockSource,
		int iInterrupt,
		ME9000MIO_P_CNT_PROC pCntIrqProc,
		void *pCntIrqContext){
	int err;

	err = cntCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntConfig()", my_errno);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntConfig()", my_errno);
		return err;
	}

	err = cntConfig(
			&boardContextVec[uiBoardNumber].cntContext,
			iRetrigger,
			iExtEvent,
			iClockSource,
			iInterrupt,
			pCntIrqProc,
			pCntIrqContext);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntConfig()", my_errno);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntConfig()", my_errno);
		return err;
	}

	return ME9000MIO_NO_ERROR;
}


int cntConfig(
		me9000mioCntContext_t *cntContext,
		int iRetrigger,
		int iExtEvent,
		int iClockSource,
		int iInterrupt,
		ME9000MIO_P_CNT_PROC pCntIrqProc,
		void *pCntIrqContext){
	int err;
	me9000mio_cnt_config_t config;
	int oflags;

	if(pCntIrqProc){
		cntContext->pCallback = pCntIrqProc;
		cntContext->pCallbackContext = pCntIrqContext;

		/* Install signal handler */
		if(signal(SIGIO, irqHandler) == SIG_ERR)
			return storeSystemError(errno);

		/* Register process as owner of the path */
		err = fcntl(cntContext->fd, F_SETOWN, getpid());
		if(err == -1)
			return storeSystemError(errno);

		/* Ask for the flags set in the path */
		oflags = fcntl(cntContext->fd, F_GETFL);
		if(oflags == -1)
			return storeSystemError(errno);

		/* Set the current process to the asynchronous queue of the driver */
		err = fcntl(cntContext->fd, F_SETFL, oflags | FASYNC);
		if(err == -1)
			return storeSystemError(errno);
	}
	else{
		cntContext->pCallback = NULL;
		cntContext->pCallbackContext = NULL;

		/* Uninstall signal handler */
		if(signal(SIGIO, SIG_DFL) == SIG_ERR)
			return storeSystemError(errno);

		/* Register process as owner of the path */
		err = fcntl(cntContext->fd, F_SETOWN, getpid());
		if(err == -1)
			return storeSystemError(errno);

		/* Ask for the flags set in the path */
		oflags = fcntl(cntContext->fd, F_GETFL);
		if(oflags == -1)
			return storeSystemError(errno);

		/* Unset the current process to the asuncronous queue of the driver */
		fcntl(cntContext->fd, F_SETFL, oflags & ~FASYNC);
		if(err == -1)
			return storeSystemError(errno);
	}

	config.retrigger = iRetrigger;
	config.ext_event = iExtEvent;
	config.interrupt = iInterrupt;
	config.clock = iClockSource;

	err = ioctl(cntContext->fd, ME9000MIO_CNT_CONFIG, &config);
	if(err) return storeSystemError(errno);

	return ME9000MIO_NO_ERROR;
}


int me9000mioCntReset(unsigned int uiBoardNumber){
	int err;

	err = cntCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntReset()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntReset()", err);
		return err;
	}

	err = cntReset(&boardContextVec[uiBoardNumber].cntContext);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntReset()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntReset()", err);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int cntReset(me9000mioCntContext_t *cntContext){
	int err;
	int i;

	for (i = 0; i < 8; i++){
		err = cntConfig(
				cntContext,
				ME9000MIO_CNT_RETRIGGER_ENABLE,
				ME9000MIO_CNT_EVENT_DISABLE,
				ME9000MIO_CNT_CLOCK_INTERNAL,
				ME9000MIO_CNT_INT_DISABLE,
				NULL,
				NULL);
		if (err) return err;

		err = cntStop(cntContext);
		if (err) return err;
	}

    return ME9000MIO_NO_ERROR;
}


int me9000mioCntWrite(unsigned int uiBoardNumber, unsigned short usValue){
	int err;

	err = cntCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntWrite()", my_errno);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntWrite()", my_errno);
		return err;
	}

	err = cntWrite(&boardContextVec[uiBoardNumber].cntContext, usValue);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntWrite()", my_errno);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntWrite()", my_errno);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int cntWrite(me9000mioCntContext_t *cntContext, unsigned short usValue){
	int err;
	uint16_t value = usValue;

	err = ioctl(cntContext->fd, ME9000MIO_CNT_WRITE, &value);
	if(err) return storeSystemError(errno);

	return ME9000MIO_NO_ERROR;
}



int me9000mioCntRead(
		unsigned int uiBoardNumber,
		unsigned short *pusValue){
	int err;

	err = cntCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntRead()", my_errno);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntRead()", my_errno);
		return err;
	}

	err = cntRead(&boardContextVec[uiBoardNumber].cntContext, pusValue);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntRead()", my_errno);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntRead()", my_errno);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int cntRead(me9000mioCntContext_t *cntContext, unsigned short *pusValue){
	int err;
	uint16_t value;

	err = ioctl(cntContext->fd, ME9000MIO_CNT_READ, &value);
	if(err) return storeSystemError(errno);

	*pusValue = value;
	return ME9000MIO_NO_ERROR;
}


int me9000mioCntStart(unsigned int uiBoardNumber){
	int err;

	err = cntCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntStart()", my_errno);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntStart()", my_errno);
		return err;
	}

	err = cntStart(&boardContextVec[uiBoardNumber].cntContext);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntStart()", my_errno);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntStart()", my_errno);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int cntStart(me9000mioCntContext_t *cntContext){
	int err;

	err = ioctl(cntContext->fd, ME9000MIO_CNT_START);
	if(err) return storeSystemError(errno);

	return ME9000MIO_NO_ERROR;
}


int me9000mioCntStop(unsigned int uiBoardNumber){
	int err;

	err = cntCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntStop()", my_errno);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntStop()", my_errno);
		return err;
	}

	err = cntStop(&boardContextVec[uiBoardNumber].cntContext);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntStop()", my_errno);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntStop()", my_errno);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int cntStop(me9000mioCntContext_t *cntContext){
	int err;

	err = ioctl(cntContext->fd, ME9000MIO_CNT_STOP);
	if(err) return storeSystemError(errno);

	return ME9000MIO_NO_ERROR;
}


int me9000mioCntGetIrqCount(
		unsigned int uiBoardNumber,
		unsigned long *pulCount){
	int err;

	err = cntCheckBoardFD(uiBoardNumber);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntGetIrqCount()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntGetIrqCount()", err);
		return err;
	}

	err = cntGetIrqCount(&boardContextVec[uiBoardNumber].cntContext, pulCount);
	if(err){
		if(errorDefaultProcFlag)
			errorDefaultProc("me9000mioCntGetIrqCount()", err);
		else if(errorUserProcPtr)
			errorUserProcPtr("me9000mioCntGetIrqCount()", err);
		return err;
	}
	return ME9000MIO_NO_ERROR;
}


int cntGetIrqCount(
		me9000mioCntContext_t *cntContext,
		unsigned long *pulCount){
	int err;

	err = ioctl(cntContext->fd, ME9000MIO_CNT_GET_IRQ_COUNT, pulCount);
	if(err) return storeSystemError(errno);

	return ME9000MIO_NO_ERROR;
}



/*===========================================================================
  Signal handler
  =========================================================================*/

void irqHandler(int sig){
	int i;

	if(sig != SIGIO) return;

	/* Call any callback routine installed */
	for(i = 0; i < board_count; i++){
		if(boardContextVec[i].cntContext.pCallback != NULL){
			boardContextVec[i].cntContext.pCallback(boardContextVec[i].cntContext.pCallbackContext);
		}
		if(boardContextVec[i].dioContext.pCallback != NULL){
			boardContextVec[i].dioContext.pCallback(boardContextVec[i].dioContext.pCallbackContext);
		}
	}

	return;
}
