/* Device driver for DIO and Counter on the Meilhaus ME-9000 board family.
 * =======================================================================
 *
 *  Copyright (C) 2004 Meilhaus Electronic GmbH (support@meilhaus.de)
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author:	Guenter Gebhardt	<g.gebhardt@meilhaus.de>
 */


/*
 * User application could also include the kernel header files. But the
 * real kernel functions are protected by #ifdef __KERNEL__.
 */
#ifndef __KERNEL__
#  define __KERNEL__
#endif


/*
 * This must be defined before module.h is included. Not needed, when
 * it is a built in driver.
 */
#ifndef MODULE
#  define MODULE
#endif


/*
 * (UTS_RELEASE, LINUX_VERSION_CODE, ...)
 */
#include <linux/version.h>


/*
 * Basic facilities for modules.
 * Defines __module_kernel_version.
 */
#include <linux/module.h>


#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/unistd.h>
#include <linux/list.h>
#include <linux/proc_fs.h>

#include <linux/poll.h>
#include <linux/vmalloc.h>
#include <asm/pgtable.h>
#include <asm/uaccess.h>
#include <linux/types.h>

#include <linux/slab.h>


/* Include-File for the Meilhaus ME-9000mio I/O board */
#include "me9000mio.h"

/*
 * Keep compatibility down to kernel 2.4
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#  error "This kernel is too old: This driver works only for kernel version 2.4 or newer"
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#  ifdef MODVERSIONS
#    include <linux/modversions.h>
#  endif

#  define USER_HZ	HZ

   typedef void irqreturn_t;
#  define IRQ_NONE
#  define IRQ_HANDLED
#  define IRQ_RETVAL(x)
#endif

/* Administrative stuff for modinfo */
MODULE_AUTHOR("Guenter Gebhardt <g.gebhardt@meilhaus.de>");
MODULE_DESCRIPTION("Device Driver Module for the DIO and Counter of the Meilhaus ME-9000 boards version 1.0.0");
MODULE_SUPPORTED_DEVICE("DIO and Counter on Meilhaus ME-9000 serial boards");


#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


/* Board specific data are kept in a global list */
LIST_HEAD(me9000mio_board_info_list);


/* Major Device Numbers. 0 means to get it automatically from the System */
static unsigned long me9000mio_dio_major_driver_no = 0;
static unsigned long me9000mio_cnt_major_driver_no = 0;


/* Let the user specify a custom major driver number */

#ifdef module_param
	module_param(me9000mio_dio_major_driver_no, long, S_IRUGO);
	module_param(me9000mio_cnt_major_driver_no, long, S_IRUGO);
#else
	MODULE_PARM(me9000mio_dio_major_driver_no, "i");
	MODULE_PARM(me9000mio_cnt_major_driver_no, "i");
#endif
MODULE_PARM_DESC(me9000mio_dio_major_driver_no, "Major driver number digital I/O (default 0)");
MODULE_PARM_DESC(me9000mio_cnt_major_driver_no, "Major driver number for counter (default 0)");


/*-----------------------------------------------------------------------------
  Module stuff
  ---------------------------------------------------------------------------*/
int init_module(void);
void cleanup_module(void);


/*-----------------------------------------------------------------------------
  Board detection and initialization
  ---------------------------------------------------------------------------*/
static int get_devices(struct pci_driver *);
static int me9000mio_probe(struct pci_dev *dev, const struct pci_device_id *id);

static void clear_board_info_list(void);
static int get_registers(struct pci_dev *dev, me9000mio_info_t *info);
static int init_board_info(struct pci_dev *dev, me9000mio_info_t *info);
static int alloc_cnt_context(me9000mio_info_t *info);
static int alloc_dio_context(me9000mio_info_t *info);


/*-----------------------------------------------------------------------------
  Stuff used by all device parts
  ---------------------------------------------------------------------------*/
static int me9000mio_open(struct inode *, struct file *);
static int me9000mio_release(struct inode *, struct file *);
static int me9000mio_read_procmem(char *, char **, off_t, int, int *, void *);


/*-----------------------------------------------------------------------------
  General stuff
  ---------------------------------------------------------------------------*/
static int me9000mio_get_user_info(me9000mio_user_info_t *, me9000mio_info_t *);


/*-----------------------------------------------------------------------------
  Digital I/O stuff
  ---------------------------------------------------------------------------*/
static int me9000mio_dio_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static int me9000mio_dio_config(me9000mio_dio_config_t *, me9000mio_dio_context_t *);
static int me9000mio_dio_get_line(me9000mio_dio_line_t *, me9000mio_dio_context_t *);
static int me9000mio_dio_set_line(me9000mio_dio_line_t *, me9000mio_dio_context_t *);
static int me9000mio_dio_get_byte(u8 *, me9000mio_dio_context_t *);
static int me9000mio_dio_set_byte(u8 *, me9000mio_dio_context_t *);
static int me9000mio_dio_tristate(me9000mio_dio_line_t *, me9000mio_dio_context_t *);
static int me9000mio_dio_fasync(int, struct file *, int);
static int me9000mio_dio_disable_int(me9000mio_dio_context_t *);
static int me9000mio_dio_get_irq_count(unsigned long *, me9000mio_dio_context_t *);
static irqreturn_t me9000mio_dio_isr(int, void *
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
, struct pt_regs *
#endif
);

/*-----------------------------------------------------------------------------
  Counter stuff
  ---------------------------------------------------------------------------*/
static int me9000mio_cnt_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static int me9000mio_cnt_config(me9000mio_cnt_config_t *, me9000mio_cnt_context_t *);
static int me9000mio_cnt_read(u16 *, me9000mio_cnt_context_t *);
static int me9000mio_cnt_write(u16 *, me9000mio_cnt_context_t *);
static int me9000mio_cnt_start(me9000mio_cnt_context_t *);
static int me9000mio_cnt_stop(me9000mio_cnt_context_t *);
static int me9000mio_cnt_fasync(int, struct file *, int);
static int me9000mio_cnt_disable_int(me9000mio_cnt_context_t *);
static int me9000mio_cnt_get_irq_count(unsigned long *, me9000mio_cnt_context_t *);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
static irqreturn_t me9000mio_cnt_isr(int, void *);
#else
static irqreturn_t me9000mio_cnt_isr(int, void *, struct pt_regs *);
#endif


/*-----------------------------------------------------------------------------
  Inline functions
  ---------------------------------------------------------------------------*/
static void inline me9000mio_outb(u8 value, unsigned long port);
static u8 inline me9000mio_inb(unsigned long port);


static void me9000mio_outb(u8 value, unsigned long port){
	PORT_PDEBUG("--> 0x%02X port 0x%04lX\n", value, port);
	writeb(value, (void *)port);
}

static u8 me9000mio_inb(unsigned long port){
	u8 value;
	value = readb((void *)port);
	PORT_PDEBUG("<-- 0x%02X port 0x%04lX\n", value, port);
	return value;
}


struct pci_driver me9000mio_driver = {
	.name = ME9000MIO_NAME,
	.id_table = me9000mio_pci_table,
	.probe = me9000mio_probe
};


static struct file_operations me9000mio_dio_fops = {
	.owner = THIS_MODULE,
	.ioctl = me9000mio_dio_ioctl,
	.open = me9000mio_open,
	.release = me9000mio_release,
	.fasync = me9000mio_dio_fasync,
};


static struct file_operations me9000mio_cnt_fops = {
	.owner = THIS_MODULE,
	.ioctl = me9000mio_cnt_ioctl,
	.open = me9000mio_open,
	.release = me9000mio_release,
	.fasync = me9000mio_cnt_fasync,
};


static int get_devices(struct pci_driver *drv){
	struct pci_dev *dev = NULL;
	int i;
	int result;
	int num = 0;

	CALL_PDEBUG("get_devices() is executed\n");

	for(i = 0; i <  NUMBER_ME9000_VERSIONS; i++)
	{
		while((dev =
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 14)
		pci_find_device(
#else
		pci_get_device(
#endif
		me9000mio_pci_table[i].vendor, me9000mio_pci_table[i].device, dev))){
			result = me9000mio_probe(dev, &me9000mio_pci_table[i]);
			if(!result) num++;
		}
	}

	return num;
}


int __init me9000mio_init_module(void){
	int result = 0;

	CALL_PDEBUG("init_module() is executed\n");

	/* I use this custom detection function because the serial driver may
	   have already occupied this device and our probe function will never
	   got invoced by the pci subsystem */
	result = get_devices(&me9000mio_driver);
	PDEBUG("init_module():%d devices detected\n", result);
	if (result < 0){
		printk(KERN_ERR"ME9000MIO:init_module():Can't register driver\n");
		goto INIT_ERROR_1;
	}

	/* Allocate major number for Digital I/O */
	result = register_chrdev(me9000mio_dio_major_driver_no, ME9000MIO_DIO_NAME, &me9000mio_dio_fops);
	if (result < 0){
		printk(KERN_ERR"ME9000MIO:init_module():Can't get DIO major no\n");
		goto INIT_ERROR_2;
	}
	else{
		me9000mio_dio_major_driver_no = result;
	}

	/* Allocate major number for counter */
	result = register_chrdev(me9000mio_cnt_major_driver_no, ME9000MIO_CNT_NAME, &me9000mio_cnt_fops);
	if (result < 0){
		printk(KERN_ERR"ME9000MIO:init_module():Can't get CNT major no\n");
		goto INIT_ERROR_3;
	}
	else{
		me9000mio_cnt_major_driver_no = result;
	}

	/* Create the /proc/me9000mio entry */
	if(!create_proc_read_entry("me9000mio", 0, NULL, me9000mio_read_procmem, NULL)){
		result = -ENODEV;
		printk(KERN_ERR"ME9000MIO:init_module():Can't create proc entry\n");
		goto INIT_ERROR_4;
	}

	return 0;

INIT_ERROR_4:
	unregister_chrdev(me9000mio_cnt_major_driver_no, ME9000MIO_CNT_NAME);

INIT_ERROR_3:
	unregister_chrdev(me9000mio_dio_major_driver_no, ME9000MIO_DIO_NAME);

INIT_ERROR_2:
	//pci_unregister_driver(&me9000mio_driver);
	clear_board_info_list();

INIT_ERROR_1:
	return result;
}
module_init(me9000mio_init_module);


static void clear_board_info_list(void){
	struct list_head *board_p;
	me9000mio_info_t *board_info;

	CALL_PDEBUG("clear_board_info_list() is executed\n");

	/* Clear the device contexts */
	for(board_p = me9000mio_board_info_list.next; board_p != &me9000mio_board_info_list; board_p = board_p->next){
		board_info = list_entry(board_p, me9000mio_info_t, list);

		/* Clear digital I/O context */
		kfree(board_info->dio_context);

		/* Clear counter context */
		kfree(board_info->cnt_context);
	}

	/* Clear the board info list */
	while(!list_empty(&me9000mio_board_info_list)){
		board_p = me9000mio_board_info_list.next;
		board_info = list_entry(board_p, me9000mio_info_t, list);
		//pci_release_regions(board_info->pci_dev_p);
		iounmap((void *) board_info->me9000mio_regbase);
		list_del(board_p);
		kfree(board_info);
	}
}


static int get_registers(struct pci_dev *dev, me9000mio_info_t *board_info){

	CALL_PDEBUG("get_registers() is executed\n");

	board_info->me9000mio_regbase = pci_resource_start (dev, 0);
	if(board_info->me9000mio_regbase == 0){
		printk(KERN_ERR"ME9000MIO:get_registers():PCI base address 0 is not available\n");
		return -ENODEV;
	}
	board_info->me9000mio_regbase_size = pci_resource_len(dev, 0);

	PDEBUG("get_registers():ME9000MIO registers at address 0x%4lX [0x%4lX]\n",
			board_info->me9000mio_regbase,
			board_info->me9000mio_regbase_size);

	if(board_info->me9000mio_regbase & PCI_BASE_ADDRESS_SPACE){
		printk(KERN_ERR"ME9000MIO:get_registers():ME9000MIO space is not MEM\n");
		return -EIO;
	}
	board_info->me9000mio_regbase =
		(unsigned long) ioremap(board_info->me9000mio_regbase & PCI_BASE_ADDRESS_MEM_MASK, board_info->me9000mio_regbase_size);
	PDEBUG("get_registers():MEM mapped to 0x%08lX\n", board_info->me9000mio_regbase);

	return 0;
}


static int init_board_info(struct pci_dev *pci_dev_p, me9000mio_info_t *board_info){
	int i;
	int result;
	struct list_head *board_p;
	board_info->pci_dev_p = pci_dev_p;

	CALL_PDEBUG("init_board_info() is executed\n");

	/* Get the index of the board in the global list */
	for(board_p = me9000mio_board_info_list.next, i = 0; board_p != &me9000mio_board_info_list; board_p = board_p->next, i++){
		if(board_p == &board_info->list){
			board_info->board_count = i;
			break;
		}
	}
	if(board_p == &me9000mio_board_info_list){
		printk(KERN_ERR"ME9000MIO:init_board_info():Cannot get index of baord\n");
		return -ENODEV;
	}

	/* Get the vendor id */
	board_info->vendor_id = pci_dev_p->vendor;
	PDEBUG("init_board_info():vendor_id = 0x%x\n", board_info->vendor_id);

	/* Get the device id */
	board_info->device_id = pci_dev_p->device;
	PDEBUG("init_board_info():device_id = 0x%x\n", board_info->device_id);

	/* Get the pci device number */
	board_info->pci_dev_no = PCI_FUNC(pci_dev_p->devfn);
	PDEBUG("init_board_info():pci_func_no = 0x%x\n", board_info->pci_func_no);

	/* Get the pci slot number */
	board_info->pci_dev_no = PCI_SLOT(pci_dev_p->devfn);
	PDEBUG("init_board_info():pci_dev_no = 0x%x\n", board_info->pci_dev_no);

	/* Get the pci bus number */
	board_info->pci_bus_no = pci_dev_p->bus->number;
	PDEBUG("init_board_info():pci_bus_no = 0x%x\n", board_info->pci_bus_no);

	/* Get the serial number */
	result = pci_read_config_dword(pci_dev_p, 0x2C, &board_info->serial_no);
	if(result != PCIBIOS_SUCCESSFUL){
		printk(KERN_ERR"ME4000:init_board_info: Can't get serial_no\n");
		return result;
	}
	PDEBUG("init_board_info():serial_no = 0x%x\n", board_info->serial_no);

	/* Get the irq assigned to the board */
	board_info->irq = pci_dev_p->irq;
	PDEBUG("init_board_info():irq = %d\n", board_info->irq);

	return 0;
}


static int alloc_dio_context(me9000mio_info_t *info){
	me9000mio_dio_context_t *dio_context;

	CALL_PDEBUG("alloc_dio_context() is executed\n");

	dio_context = kmalloc(sizeof(me9000mio_dio_context_t), GFP_KERNEL);
	if(!dio_context){
		printk(KERN_ERR"ME9000MIO:alloc_dio_context():Can't get memory for dio context\n");
		return -ENOMEM;
	}
	memset(dio_context, 0, sizeof(me9000mio_dio_context_t));

	info->dio_context = dio_context;

	spin_lock_init(&dio_context->use_lock);
	dio_context->board_info = info;

	dio_context->int_reg = info->me9000mio_regbase + ME9000MIO_MPIO_INT_REG;
	dio_context->lvl_reg = info->me9000mio_regbase + ME9000MIO_MPIO_LVL_REG;
	dio_context->threestate_reg = info->me9000mio_regbase + ME9000MIO_MPIO_3T_REG;
	dio_context->inv_reg = info->me9000mio_regbase + ME9000MIO_MPIO_INV_REG;
	dio_context->sel_reg = info->me9000mio_regbase + ME9000MIO_MPIO_SEL_REG;
	dio_context->irq_status_reg = info->me9000mio_regbase + ME9000MIO_IRQ_STATUS_REG;
	dio_context->irq_source_reg = info->me9000mio_regbase + ME9000MIO_IRQ_SOURCE_REG;

	dio_context->irq = info->irq;
	dio_context->fasync_ptr = NULL;

	return 0;
}


static int alloc_cnt_context(me9000mio_info_t *info){
	me9000mio_cnt_context_t *cnt_context;

	CALL_PDEBUG("alloc_cnt_context() is executed\n");

	cnt_context = kmalloc(sizeof(me9000mio_cnt_context_t), GFP_KERNEL);
	if(!cnt_context){
		printk(KERN_ERR"ME9000MIO:alloc_cnt_context():Can't get memory for cnt context\n");
		return -ENOMEM;
	}
	memset(cnt_context, 0, sizeof(me9000mio_cnt_context_t));

	info->cnt_context = cnt_context;

	spin_lock_init(&cnt_context->use_lock);
	cnt_context->board_info = info;

	cnt_context->cntl_reg = info->me9000mio_regbase + ME9000MIO_CNT_CNTL_REG;
	cnt_context->lsb_reg = info->me9000mio_regbase + ME9000MIO_CNT_LSB_REG;
	cnt_context->msb_reg = info->me9000mio_regbase + ME9000MIO_CNT_MSB_REG;
	cnt_context->irq_status_reg = info->me9000mio_regbase + ME9000MIO_IRQ_STATUS_REG;
	cnt_context->irq_source_reg = info->me9000mio_regbase + ME9000MIO_IRQ_SOURCE_REG;

	cnt_context->irq = info->irq;
	cnt_context->fasync_ptr = NULL;

	return 0;
}


static int me9000mio_probe(struct pci_dev *dev, const struct pci_device_id *id){
	int result = 0;
	me9000mio_info_t *board_info;

	CALL_PDEBUG("me9000mio_probe() is executed\n");

	/* Allocate structure for board context */
	board_info = kmalloc(sizeof(me9000mio_info_t), GFP_KERNEL);
	if(!board_info){
		printk(KERN_ERR"ME9000MIO:Can't get memory for board info structure\n");
		result = -ENOMEM;
		goto PROBE_ERROR_1;
	}
	memset(board_info, 0, sizeof(me9000mio_info_t));

	/* Add to global linked list */
	list_add_tail(&board_info->list, &me9000mio_board_info_list);

	/* Get the PCI base registers */
	result = get_registers(dev, board_info);
	if(result){
		printk(KERN_ERR"me9000mio_probe():Cannot get registers\n");
		goto PROBE_ERROR_2;
	}

	/* Request the PCI register regions */
	/*
	result = pci_request_regions(dev, ME9000MIO_NAME);
	if (result < 0){
		printk(KERN_ERR"me9000mio_probe():Cannot request I/O regions\n");
		goto PROBE_ERROR_2;
	}
	*/

	/* Initialize board info */
	result = init_board_info(dev, board_info);
	if (result){
		printk(KERN_ERR"me9000mio_probe():Cannot init baord info\n");
		goto PROBE_ERROR_3;
	}

	/* Allocate digital I/O context */
	result = alloc_dio_context(board_info);
	if (result){
		printk(KERN_ERR"me9000mio_probe():Cannot allocate dio context\n");
		goto PROBE_ERROR_3;
	}

	/* Allocate counter context */
	result = alloc_cnt_context(board_info);
	if (result){
		printk(KERN_ERR"me9000mio_probe():Cannot allocate cnt context\n");
		goto PROBE_ERROR_4;
	}

	return 0;

PROBE_ERROR_4:
	kfree(board_info->dio_context);

PROBE_ERROR_3:
	//pci_release_regions(dev);

PROBE_ERROR_2:
	list_del(&board_info->list);
	kfree(board_info);

PROBE_ERROR_1:
	return result;
}


static int me9000mio_open(struct inode *inode_p, struct file *file_p){
	int board;
	int err = 0;
	struct list_head *ptr;
	me9000mio_info_t *board_info = NULL;
	me9000mio_dio_context_t *dio_context = NULL;
	me9000mio_cnt_context_t *cnt_context = NULL;

	CALL_PDEBUG("me9000mio_open() is executed\n");

	/* Digital I/O */
	if(MAJOR(inode_p->i_rdev) == me9000mio_dio_major_driver_no){
		board = DIO_BOARD(inode_p->i_rdev);

		PDEBUG("me9000mio_open():DIO board = %d\n", board);

		/* Search for the board context */
		for(ptr = me9000mio_board_info_list.next; ptr != &me9000mio_board_info_list; ptr = ptr->next){
			board_info = list_entry(ptr, me9000mio_info_t, list);
			if(board_info->board_count == board) break;
		}

		if(ptr == &me9000mio_board_info_list){
			printk(KERN_ERR"ME9000MIO:me9000mio_open():DIO board %d not in device list\n", board);
			return -ENODEV;
		}

		/* Search for the dio context */
		dio_context = board_info->dio_context;

		/* Check if already opened */
		spin_lock(&dio_context->use_lock);
		if(dio_context->in_use){
			printk(KERN_ERR"ME9000MIO:me9000mio_open():DIO already in use\n");
			spin_unlock(&dio_context->use_lock);
			return -EBUSY;
		}
		dio_context->in_use = 1;
		spin_unlock(&dio_context->use_lock);

		/* Hold the context in private data */
		file_p->private_data = dio_context;

		/* Set file operations pointer to single functions */
		file_p->f_op = &me9000mio_dio_fops;

		/* Request the interrupt line */
		err = request_irq(dio_context->irq, me9000mio_dio_isr,
#ifdef IRQF_DISABLED
					IRQF_DISABLED | IRQF_SHARED,
#else
					SA_INTERRUPT | SA_SHIRQ,
#endif
					ME9000MIO_NAME, dio_context);
		if(err){
			printk(KERN_ERR"ME9000MIO:me9000mio_open():Can't get DIO interrupt line err=%d\n", err);
			dio_context->in_use = 0;
			return -ENODEV;
		}
	}
	/* Counters */
	else if(MAJOR(inode_p->i_rdev) == me9000mio_cnt_major_driver_no){
		board = CNT_BOARD(inode_p->i_rdev);

		PDEBUG("me9000mio_open():CNT board = %d\n", board);

		/* Search for the board context */
		for(ptr = me9000mio_board_info_list.next; ptr != &me9000mio_board_info_list; ptr = ptr->next){
			board_info = list_entry(ptr, me9000mio_info_t, list);
			if(board_info->board_count == board) break;
		}

		if(ptr == &me9000mio_board_info_list){
			printk(KERN_ERR"ME9000MIO:me9000mio_open():CNT board %d not in device list\n", board);
			return -ENODEV;
		}

		/* Get the cnt context */
		cnt_context = board_info->cnt_context;

		/* Check if already opened */
		spin_lock(&cnt_context->use_lock);
		if(cnt_context->in_use){
			printk(KERN_ERR"ME9000MIO:me9000mio_open():CNT already in use\n");
			spin_unlock(&cnt_context->use_lock);
			return -EBUSY;
		}
		cnt_context->in_use = 1;
		spin_unlock(&cnt_context->use_lock);

		/* Hold the context in private data */
		file_p->private_data = cnt_context;

		/* Set file operations pointer to single functions */
		file_p->f_op = &me9000mio_cnt_fops;

		/* Request the interrupt line */
		err = request_irq(cnt_context->irq, me9000mio_cnt_isr,
#ifdef IRQF_DISABLED
					IRQF_DISABLED | IRQF_SHARED,
#else
					SA_INTERRUPT | SA_SHIRQ,
#endif
					ME9000MIO_NAME, cnt_context);
		if(err){
			printk(KERN_ERR"ME9000MIO:me9000mio_open():Can't get CNT interrupt line");
			dio_context->in_use = 0;
			return -ENODEV;
		}
	}
	else {
		printk(KERN_ERR"ME9000MIO:me9000mio_open():Major number unknown\n");
		return -EINVAL;
	}

	return 0;
}


static int me9000mio_release(struct inode *inode_p, struct file *file_p){
	me9000mio_dio_context_t *dio_context;
	me9000mio_cnt_context_t *cnt_context;

	CALL_PDEBUG("me9000mio_release() is executed\n");

	if(MAJOR(inode_p->i_rdev) == me9000mio_dio_major_driver_no){
		dio_context = file_p->private_data;

		/* Disable the interrupt */
		me9000mio_dio_disable_int(dio_context);

		free_irq(dio_context->irq, dio_context);

		/* Delete the fasync structure and free memory */
		me9000mio_dio_fasync(0, file_p, 0);

		/* Mark digital I/O as unused */
		dio_context->in_use = 0;
	}
	else if(MAJOR(inode_p->i_rdev) == me9000mio_cnt_major_driver_no){
		cnt_context = file_p->private_data;

		/* Disable the interrupt */
		me9000mio_cnt_disable_int(cnt_context);

		free_irq(cnt_context->irq, cnt_context);

		/* Delete the fasync structure and free memory */
		me9000mio_cnt_fasync(0, file_p, 0);

		/* Mark counters as unused */
		cnt_context->in_use = 0;
	}
	else {
		printk(KERN_ERR"ME9000MIO:me9000mio_release():Major number unknown\n");
		return -EINVAL;
	}

	return 0;
}



/*------------------------------------ General stuff ----------------------------------------------*/

static int me9000mio_get_user_info(me9000mio_user_info_t *arg, me9000mio_info_t *info){
	CALL_PDEBUG("me9000mio_get_user_info() is executed\n");

	arg->board_count = info->board_count;
	arg->me9000mio_regbase = info->me9000mio_regbase;
	arg->vendor_id = info->vendor_id;
	arg->device_id = info->device_id;
	arg->serial_no = info->serial_no;
	arg->pci_bus_no = info->pci_bus_no;
	arg->pci_dev_no = info->pci_dev_no;
	arg->pci_func_no = info->pci_func_no;
	arg->irq = info->irq;
	arg->driver_version = ME9000MIO_DRIVER_VERSION;

	return 0;
}



/*------------------------------------ DIO stuff ----------------------------------------------*/

static int me9000mio_dio_ioctl(struct inode * inode_p, struct file *file_p, unsigned int service, unsigned long arg){
	me9000mio_dio_context_t *dio_context;

	CALL_PDEBUG("me9000mio_dio_ioctl() is executed\n");

	dio_context = file_p->private_data;

	if(_IOC_TYPE(service) != ME9000MIO_MAGIC){
		printk(KERN_ERR"me9000mio_dio_ioctl():Wrong magic number\n");
		return -ENOTTY;
	}

	switch(service){
		case ME9000MIO_GET_USER_INFO:
			return me9000mio_get_user_info((me9000mio_user_info_t *) arg, dio_context->board_info);
		case ME9000MIO_DIO_CONFIG:
			return me9000mio_dio_config((me9000mio_dio_config_t *) arg, dio_context);
		case ME9000MIO_DIO_GET_LINE:
			return me9000mio_dio_get_line((me9000mio_dio_line_t *) arg, dio_context);
		case ME9000MIO_DIO_SET_LINE:
			return me9000mio_dio_set_line((me9000mio_dio_line_t *) arg, dio_context);
		case ME9000MIO_DIO_GET_BYTE:
			return me9000mio_dio_get_byte((u8 *) arg, dio_context);
		case ME9000MIO_DIO_SET_BYTE:
			return me9000mio_dio_set_byte((u8 *) arg, dio_context);
		case ME9000MIO_DIO_TRISTATE:
			return me9000mio_dio_tristate((me9000mio_dio_line_t *) arg, dio_context);
		case ME9000MIO_DIO_GET_IRQ_COUNT:
			return me9000mio_dio_get_irq_count((unsigned long *) arg, dio_context);
		default:
			printk(KERN_ERR"ME9000MIO:me9000mio_dio_ioctl():Invalid service number %d\n", service);
			return -ENOTTY;
	}

	return 0;
}


static int me9000mio_dio_config(me9000mio_dio_config_t *arg, me9000mio_dio_context_t *dio_context){
	me9000mio_dio_config_t cmd;
	int err;
	u8 tmp;
	u8 mask;

	CALL_PDEBUG("me9000mio_dio_config() is executed\n");

	/* Copy data from user */
	err = copy_from_user(&cmd, arg, sizeof(me9000mio_dio_config_t));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_config():Can't copy from user space\n");
		return -EFAULT;
	}

	/* Check parameters */
	if(cmd.line > 7){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_config():Line %d is not available\n", cmd.line);
		return -EINVAL;
	}

	mask = 1 << cmd.line;

	/* Disable interrupt for sure */
	tmp = me9000mio_inb(dio_context->int_reg);
	tmp &= ~mask;
	me9000mio_outb(tmp, dio_context->int_reg);

	switch(cmd.dir){
		case ME9000MIO_DIO_OUTPUT:
			tmp = me9000mio_inb(dio_context->sel_reg);
			tmp &= ~mask;
			me9000mio_outb(tmp, dio_context->sel_reg);
			break;
		case ME9000MIO_DIO_INPUT:
			tmp = me9000mio_inb(dio_context->sel_reg);
			tmp |= mask;
			me9000mio_outb(tmp, dio_context->sel_reg);
			break;
		default:
			printk(KERN_ERR"ME9000MIO:me9000mio_dio_config():Invalid line direction\n");
			return -EINVAL;
	}

	switch(cmd.invert){
		case ME9000MIO_DIO_INV_DISABLE:
			tmp = me9000mio_inb(dio_context->inv_reg);
			tmp &= ~mask;
			me9000mio_outb(tmp, dio_context->inv_reg);
			break;
		case ME9000MIO_DIO_INV_ENABLE:
			tmp = me9000mio_inb(dio_context->inv_reg);
			tmp |= mask;
			me9000mio_outb(tmp, dio_context->inv_reg);
			break;
		default:
			printk(KERN_ERR"ME9000MIO:me9000mio_dio_config():Invalid invert mode\n");
			return -EINVAL;
	}

	switch(cmd.interrupt){
		case ME9000MIO_DIO_INT_DISABLE:
			tmp = me9000mio_inb(dio_context->int_reg);
			tmp &= ~mask;
			me9000mio_outb(tmp, dio_context->int_reg);
			break;
		case ME9000MIO_DIO_INT_ENABLE:
			tmp = me9000mio_inb(dio_context->int_reg);
			tmp |= mask;
			me9000mio_outb(tmp, dio_context->int_reg);
			break;
		default:
			printk(KERN_ERR"ME9000MIO:me9000mio_dio_config():Invalid interrupt mode\n");
			return -EINVAL;
	}

	return 0;
}


static int me9000mio_dio_get_line(me9000mio_dio_line_t *arg, me9000mio_dio_context_t *dio_context){
	me9000mio_dio_line_t cmd;
	int err;
	u8 tmp;
	u8 mask;

	CALL_PDEBUG("me9000mio_dio_get_line() is executed\n");

	/* Copy data from user */
	err = copy_from_user(&cmd, arg, sizeof(me9000mio_dio_line_t));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_get_line():Can't copy from user space\n");
		return -EFAULT;
	}

	/* Check parameters */
	if(cmd.line > 7){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_get_line():Line %d is not available\n", cmd.line);
		return -EINVAL;
	}

	/* Get the bit status */
	mask = 1 << cmd.line;
	tmp = me9000mio_inb(dio_context->lvl_reg);
	cmd.status = (tmp & mask) ? 1 : 0;

	/* Copy result back to user */
	err = copy_to_user(arg, &cmd, sizeof(me9000mio_dio_line_t));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_get_line():Can't copy to user space\n");
		return -EFAULT;
	}

	return 0;
}


static int me9000mio_dio_set_line(me9000mio_dio_line_t *arg, me9000mio_dio_context_t *dio_context){
	me9000mio_dio_line_t cmd;
	int err;
	u8 tmp;
	u8 mask;

	CALL_PDEBUG("me9000mio_dio_set_line() is executed\n");

	/* Copy data from user */
	err = copy_from_user(&cmd, arg, sizeof(me9000mio_dio_line_t));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_set_line():Can't copy from user space\n");
		return -EFAULT;
	}

	/* Check parameters */
	if(cmd.line > 7){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_set_line():Line %d is not available\n", cmd.line);
		return -EINVAL;
	}

	mask = 1 << cmd.line;

	if(me9000mio_inb(dio_context->sel_reg) & mask){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_set_line():Line %d is not in output mode\n", cmd.line);
		return -EINVAL;
	}

	/* Set the bit status */
	tmp = me9000mio_inb(dio_context->lvl_reg);
	if(cmd.status)
		tmp |= mask;
	else
		tmp &= ~mask;
	me9000mio_outb(tmp, dio_context->lvl_reg);

	return 0;
}


static int me9000mio_dio_get_byte(u8 *arg, me9000mio_dio_context_t *dio_context){
	u8 cmd;
	int err;

	CALL_PDEBUG("me9000mio_dio_get_byte() is executed\n");

	cmd = me9000mio_inb(dio_context->lvl_reg);

	/* Copy result back to user */
	err = copy_to_user(arg, &cmd, sizeof(u8));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_get_byte():Can't copy to user space\n");
		return -EFAULT;
	}

	return 0;
}


static int me9000mio_dio_set_byte(u8 *arg, me9000mio_dio_context_t *dio_context){
	u8 cmd;
	int err;

	CALL_PDEBUG("me9000mio_dio_set_byte() is executed\n");

	/* Copy data from user */
	err = copy_from_user(&cmd, arg, sizeof(u8));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_set_byte():Can't copy from user space\n");
		return -EFAULT;
	}

	me9000mio_outb(cmd, dio_context->lvl_reg);
	return 0;
}


static int me9000mio_dio_tristate(me9000mio_dio_line_t *arg, me9000mio_dio_context_t *dio_context){
	me9000mio_dio_line_t cmd;
	int err;
	u8 tmp;
	u8 mask;

	CALL_PDEBUG("me9000mio_dio_tristate() is executed\n");

	/* Copy data from user */
	err = copy_from_user(&cmd, arg, sizeof(me9000mio_dio_line_t));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_tristate():Can't copy from user space\n");
		return -EFAULT;
	}

	/* Check parameters */
	if(cmd.line > 7){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_tristate():Line %d is not available\n", cmd.line);
		return -EINVAL;
	}

	mask = 1 << cmd.line;

	switch(cmd.status){
		case ME9000MIO_DIO_OUTPUT_ENABLE:
			tmp = me9000mio_inb(dio_context->threestate_reg);
			tmp &= ~mask;
			me9000mio_outb(tmp, dio_context->threestate_reg);
			break;
		case ME9000MIO_DIO_OUTPUT_DISABLE:
			tmp = me9000mio_inb(dio_context->threestate_reg);
			tmp |= mask;
			me9000mio_outb(tmp, dio_context->threestate_reg);
			break;
		default:
			printk(KERN_ERR"ME9000MIO:me9000mio_dio_tristate():Invalid line status selected\n");
			return -EINVAL;
	}

	return 0;
}


static int me9000mio_dio_disable_int(me9000mio_dio_context_t *dio_context){

	CALL_PDEBUG("me9000mio_dio_disable_int() is executed\n");

	/* Disable all interrupts */
	me9000mio_outb(0x0, dio_context->int_reg);
	return 0;
}


static int me9000mio_dio_get_irq_count(unsigned long *arg, me9000mio_dio_context_t *dio_context){
	int err;

	CALL_PDEBUG("me9000mio_dio_get_irq_count() is executed\n");

	/* Copy result back to user */
	err = copy_to_user(arg, &dio_context->irq_count, sizeof(unsigned long));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_dio_get_irq_count():Can't copy to user space\n");
		return -EFAULT;
	}

	return 0;
}


static int me9000mio_dio_fasync(int fd, struct file *file_ptr, int mode){
	int result = 0;
	me9000mio_dio_context_t *dio_context;

	CALL_PDEBUG("me9000mio_dio_fasync() is executed\n");

	dio_context = file_ptr->private_data;
	result = fasync_helper(fd, file_ptr, mode, &dio_context->fasync_ptr);
	return result;
}


static irqreturn_t me9000mio_dio_isr(int irq, void *dev_id
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
, struct pt_regs *
#endif
)
{
	me9000mio_dio_context_t *dio_context;

	ISR_PDEBUG("me9000mio_dio_isr() is executed\n");

	dio_context = dev_id;

	/* Check if irq number is right */
	if(irq != dio_context->irq){
		ISR_PDEBUG("me9000mio_dio_isr():incorrect interrupt num: %d\n", irq);
		return IRQ_NONE;
	}

	if((me9000mio_inb(dio_context->irq_source_reg) & ME9000MIO_IRQ_SOURCE_MASK) == ME9000MIO_IRQ_SOURCE_MPIO){
		me9000mio_inb(dio_context->lvl_reg); // Clear the interrupt
		++dio_context->irq_count;
		if(dio_context->fasync_ptr){
			ISR_PDEBUG("me9000mio_dio_isr():Send signal to process\n");
			kill_fasync(&dio_context->fasync_ptr, SIGIO, POLL_IN);
		}
	}
	else {
		ISR_PDEBUG("me9000mio_dio_isr():Not this board\n");
		return IRQ_NONE;
	}

	return IRQ_HANDLED;
}


/*------------------------------------ CNT STUFF ------------------------------------*/

static int me9000mio_cnt_ioctl(struct inode * inode_p, struct file *file_p, unsigned int service, unsigned long arg){
	me9000mio_cnt_context_t *cnt_context;

	CALL_PDEBUG("me9000mio_cnt_ioctl() is executed\n");

	cnt_context = file_p->private_data;

	if(_IOC_TYPE(service) != ME9000MIO_MAGIC){
		printk(KERN_ERR"me9000mio_cnt_ioctl():Wrong magic number\n");
		return -ENOTTY;
	}

	switch(service){
		case ME9000MIO_GET_USER_INFO:
			return me9000mio_get_user_info((me9000mio_user_info_t *) arg, cnt_context->board_info);
		case ME9000MIO_CNT_CONFIG:
			return me9000mio_cnt_config((me9000mio_cnt_config_t *) arg, cnt_context);
		case ME9000MIO_CNT_READ:
			return me9000mio_cnt_read((u16 *) arg, cnt_context);
		case ME9000MIO_CNT_WRITE:
			return me9000mio_cnt_write((u16 *) arg, cnt_context);
		case ME9000MIO_CNT_START:
			return me9000mio_cnt_start(cnt_context);
		case ME9000MIO_CNT_STOP:
			return me9000mio_cnt_stop(cnt_context);
		case ME9000MIO_CNT_GET_IRQ_COUNT:
			return me9000mio_cnt_get_irq_count((unsigned long *) arg, cnt_context);
		default:
			printk(KERN_ERR"ME9000MIO:me9000mio_cnt_ioctl():Invalid service number %d\n", service);
			return -ENOTTY;
	}
	return 0;
}


static int me9000mio_cnt_config(me9000mio_cnt_config_t *arg, me9000mio_cnt_context_t *cnt_context){
	me9000mio_cnt_config_t cmd;
	u8 cntl = 0;
	int err;

	CALL_PDEBUG("me9000mio_cnt_config() is executed\n");

	/* Copy data from user */
	err = copy_from_user(&cmd, arg, sizeof(me9000mio_cnt_config_t));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_cnt_config():Can't copy from user space\n");
		return -EFAULT;
	}

	/* Check retrigger parameter */
	switch(cmd.retrigger){
		case ME9000MIO_CNT_RETRIGGER_DISABLE:
			cntl |= ME9000MIO_CNT_CNTL_ONE_SHOT_BIT;
			break;
		case ME9000MIO_CNT_RETRIGGER_ENABLE:
			break;
		default:
			printk(KERN_ERR"ME9000MIO:me9000mio_cnt_config():Invalid retrigger setting\n");
			return -EINVAL;
	}

	/* Check ext_event parameter */
	switch(cmd.ext_event){
		case ME9000MIO_CNT_EVENT_DISABLE:
			break;
		case ME9000MIO_CNT_EVENT_ENABLE:
			cntl |= ME9000MIO_CNT_CNTL_EXT_EVENT_BIT;
			break;
		default:
			printk(KERN_ERR"ME9000MIO:me9000mio_cnt_config():Invalid ext_event setting\n");
			return -EINVAL;
	}

	/* Check interrupt parameter */
	switch(cmd.interrupt){
		case ME9000MIO_CNT_INT_DISABLE:
			break;
		case ME9000MIO_CNT_INT_ENABLE:
			cntl |= ME9000MIO_CNT_CNTL_INT_ENABLE_BIT;
			break;
		default:
			printk(KERN_ERR"ME9000MIO:me9000mio_cnt_config():Invalid interrupt setting\n");
			return -EINVAL;
	}

	/* Check ext_clock parameter */
	switch(cmd.clock){
		case ME9000MIO_CNT_CLOCK_INTERNAL:
			break;
		case ME9000MIO_CNT_CLOCK_EXTERNAL:
			cntl |= ME9000MIO_CNT_CNTL_EXT_CLOCK_BIT;
			break;
		default:
			printk(KERN_ERR"ME9000MIO:me9000mio_cnt_config():Invalid ext_clock setting\n");
			return -EINVAL;
	}

	/* Write the control word */
	me9000mio_outb(cntl, cnt_context->cntl_reg);

	return 0;
}


static int me9000mio_cnt_read(u16 *arg, me9000mio_cnt_context_t *cnt_context){
	u16 tmp;
	int err;

	CALL_PDEBUG("me9000mio_cnt_read() is executed\n");

	tmp = me9000mio_inb(cnt_context->lsb_reg) | (me9000mio_inb(cnt_context->msb_reg) << 8);

	/* Copy result back to user */
	err = copy_to_user(arg, &tmp, sizeof(u16));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_cnt_read():Can't copy to user space\n");
		return -EFAULT;
	}

	return 0;
}


static int me9000mio_cnt_write(u16 *arg, me9000mio_cnt_context_t *cnt_context){
	u16 cmd;
	int err;

	CALL_PDEBUG("me9000mio_cnt_write() is executed\n");

	/* Copy data from user */
	err = copy_from_user(&cmd, arg, sizeof(u16));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_cnt_write():Can't copy from user space\n");
		return -EFAULT;
	}

	/* Write to counter */
	me9000mio_outb(cmd & 0xFF, cnt_context->lsb_reg);
	me9000mio_outb((cmd >> 8) & 0xFF, cnt_context->msb_reg);

	return 0;
}


static int me9000mio_cnt_start(me9000mio_cnt_context_t *cnt_context){
	u8 tmp;

	CALL_PDEBUG("me9000mio_cnt_start() is executed\n");

	/* Start counter */
	tmp = me9000mio_inb(cnt_context->cntl_reg);
	tmp |= ME9000MIO_CNT_CNTL_START_BIT;
	me9000mio_outb(tmp, cnt_context->cntl_reg);

	return 0;
}


static int me9000mio_cnt_stop(me9000mio_cnt_context_t *cnt_context){
	u8 tmp;

	CALL_PDEBUG("me9000mio_cnt_stop() is executed\n");

	/* Stop counter */
	tmp = me9000mio_inb(cnt_context->cntl_reg);
	tmp &= ~ME9000MIO_CNT_CNTL_START_BIT;
	me9000mio_outb(tmp, cnt_context->cntl_reg);

	return 0;
}


static int me9000mio_cnt_get_irq_count(unsigned long *arg, me9000mio_cnt_context_t *cnt_context){
	int err;

	CALL_PDEBUG("me9000mio_cnt_get_irq_count() is executed\n");

	/* Copy result back to user */
	err = copy_to_user(arg, &cnt_context->irq_count, sizeof(unsigned long));
	if(err){
		printk(KERN_ERR"ME9000MIO:me9000mio_cnt_get_irq_count():Can't copy to user space\n");
		return -EFAULT;
	}

	return 0;
}


static int me9000mio_cnt_fasync(int fd, struct file *file_ptr, int mode){
	int result = 0;
	me9000mio_cnt_context_t *cnt_context;

	CALL_PDEBUG("me9000mio_cnt_fasync() is executed\n");

	cnt_context = file_ptr->private_data;
	result = fasync_helper(fd, file_ptr, mode, &cnt_context->fasync_ptr);
	return result;
}


static int me9000mio_cnt_disable_int(me9000mio_cnt_context_t *cnt_context){
	u8 tmp;

	CALL_PDEBUG("me9000mio_cnt_disable_int() is executed\n");

	/* Disable the interrupt */
	tmp = me9000mio_inb(cnt_context->cntl_reg);
	tmp &= ~ME9000MIO_CNT_CNTL_INT_ENABLE_BIT;
	me9000mio_outb(tmp, cnt_context->cntl_reg);
	return 0;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
static irqreturn_t me9000mio_cnt_isr(int irq, void *dev_id)
#else
static irqreturn_t me9000mio_cnt_isr(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
	me9000mio_cnt_context_t *cnt_context;

	ISR_PDEBUG("me9000mio_cnt_isr() is executed\n");

	cnt_context = dev_id;

	/* Check if irq number is right */
	if(irq != cnt_context->irq){
		ISR_PDEBUG("me9000mio_cnt_isr():incorrect interrupt num: %d\n", irq);
		return IRQ_NONE;
	}

	if((me9000mio_inb(cnt_context->irq_source_reg) & ME9000MIO_IRQ_SOURCE_MASK) == ME9000MIO_IRQ_SOURCE_CNT){
		me9000mio_inb(cnt_context->cntl_reg); // Clear the interrupt
		++cnt_context->irq_count;
		if(cnt_context->fasync_ptr){
			ISR_PDEBUG("me9000mio_cnt_isr():Send signal to process\n");
			kill_fasync(&cnt_context->fasync_ptr, SIGIO, POLL_IN);
		}
	}
	else {
		ISR_PDEBUG("me9000mio_cnt_isr():Not this board\n");
		return IRQ_NONE;
	}

	return IRQ_HANDLED;
}


void __exit me9000mio_module_exit(void){

	CALL_PDEBUG("cleanup_module() is executed\n");

	unregister_chrdev(me9000mio_cnt_major_driver_no, ME9000MIO_CNT_NAME);
	unregister_chrdev(me9000mio_dio_major_driver_no, ME9000MIO_DIO_NAME);
	remove_proc_entry("me9000mio", NULL);

	//pci_unregister_driver(&me9000mio_driver);
	clear_board_info_list();
}


module_exit(me9000mio_module_exit);


static int me9000mio_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	int len = 0;
	int limit = count - 1000;
	me9000mio_info_t *board_info;
	struct list_head *ptr;

	len += sprintf(buf + len, "ME9000MIO DRIVER VERSION %X.%X.%X\n\n",
			(ME9000MIO_DRIVER_VERSION & 0xFF0000) >> 16,
			(ME9000MIO_DRIVER_VERSION & 0xFF00) >> 8,
			(ME9000MIO_DRIVER_VERSION & 0xFF));

	/* Search for the board context */
	for(ptr = me9000mio_board_info_list.next; (ptr != &me9000mio_board_info_list) && (len < limit); ptr = ptr->next){
		board_info = list_entry(ptr, me9000mio_info_t, list);

		len += sprintf(buf + len, "Board number %d:\n", board_info->board_count);
		len += sprintf(buf + len, "---------------\n");
		len += sprintf(buf + len, "ME9000MIO base register = 0x%lX\n", board_info->me9000mio_regbase);
		len += sprintf(buf + len, "ME9000MIO base register size = 0x%lX\n", board_info->me9000mio_regbase_size);
		len += sprintf(buf + len, "Vendor id = 0x%X\n", board_info->vendor_id);
		len += sprintf(buf + len, "Device id = 0x%X\n", board_info->device_id);
		len += sprintf(buf + len, "Serial number = 0x%08X\n", board_info->serial_no);
		len += sprintf(buf + len, "PCI bus number = %d\n", board_info->pci_bus_no);
		len += sprintf(buf + len, "PCI device number = %d\n", board_info->pci_dev_no);
		len += sprintf(buf + len, "PCI function number = %d\n", board_info->pci_func_no);
		len += sprintf(buf + len, "IRQ = %d\n\n", board_info->irq);
	}

	*eof = 1;
	return len;
}
